"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateApi = void 0;
const typescript_1 = __importDefault(require("typescript"));
const path = __importStar(require("path"));
const lodash_1 = require("lodash");
const generate_1 = __importStar(require("@kahirokunn/oazapfts-patched/lib/codegen/generate"));
const tscodegen_1 = require("@kahirokunn/oazapfts-patched/lib/codegen/tscodegen");
const utils_1 = require("./utils");
const codegen_1 = require("./codegen");
const factory_1 = require("./utils/factory");
function defaultIsDataResponse(code) {
    const parsedCode = Number(code);
    return !Number.isNaN(parsedCode) && parsedCode >= 200 && parsedCode < 300;
}
function getOperationName({ verb, path, operation }) {
    return (0, generate_1.getOperationName)(verb, path, operation.operationId);
}
function patternMatches(pattern) {
    const filters = Array.isArray(pattern) ? pattern : [pattern];
    return function matcher(operationName) {
        if (!pattern)
            return true;
        return filters.some((filter) => typeof filter === 'string' ? filter === operationName : filter === null || filter === void 0 ? void 0 : filter.test(operationName));
    };
}
function operationMatches(pattern) {
    const checkMatch = typeof pattern === 'function' ? pattern : patternMatches(pattern);
    return function matcher(operationDefinition) {
        if (!pattern)
            return true;
        const operationName = getOperationName(operationDefinition);
        return checkMatch(operationName, operationDefinition);
    };
}
function withQueryComment(node, def, hasTrailingNewLine) {
    const comment = def.origin === 'param' ? def.param.description : def.body.description;
    if (comment) {
        return typescript_1.default.addSyntheticLeadingComment(node, typescript_1.default.SyntaxKind.MultiLineCommentTrivia, `* ${comment} `, hasTrailingNewLine);
    }
    return node;
}
async function generateApi(spec, { apiFile, apiImport = 'apiClient', optionsImport = 'Options', argSuffix = 'ApiArg', responseSuffix = 'ApiResponse', outputFile, isDataResponse = defaultIsDataResponse, filterEndpoints, unionUndefined = true, strict = true, }) {
    const v3Doc = await (0, utils_1.getV3Doc)(spec);
    const apiGen = new generate_1.default(v3Doc, {
        unionUndefined,
    });
    const operationDefinitions = (0, utils_1.getOperationDefinitions)(v3Doc).filter(operationMatches(filterEndpoints));
    const resultFile = typescript_1.default.createSourceFile('someFileName.ts', '', typescript_1.default.ScriptTarget.Latest, 
    /*setParentNodes*/ false, typescript_1.default.ScriptKind.TS);
    const printer = typescript_1.default.createPrinter({ newLine: typescript_1.default.NewLineKind.LineFeed });
    const interfaces = {};
    function registerInterface(declaration) {
        const name = declaration.name.escapedText.toString();
        if (name in interfaces && strict) {
            throw new Error(`interface/type alias ${name} already registered`);
        }
        interfaces[name] = declaration;
        return declaration;
    }
    if (outputFile) {
        outputFile = path.resolve(process.cwd(), outputFile);
        if (apiFile.startsWith('.')) {
            apiFile = path.relative(path.dirname(outputFile), apiFile);
            apiFile = apiFile.replace(/\\/g, '/');
            if (!apiFile.startsWith('.'))
                apiFile = './' + apiFile;
        }
    }
    apiFile = apiFile.replace(/\.[jt]sx?$/, '');
    const sourceCode = printer.printNode(typescript_1.default.EmitHint.Unspecified, factory_1.factory.createSourceFile([
        (0, codegen_1.generateImportNode)(apiFile, {
            [apiImport]: { isTypeOnly: false, name: 'apiClient' },
            [optionsImport]: { isTypeOnly: true, name: 'Options' },
        }),
        ...operationDefinitions.map((operationDefinition) => generateRequester({
            operationDefinition,
        })),
        ...Object.values(interfaces),
        ...apiGen['aliases'],
    ], factory_1.factory.createToken(typescript_1.default.SyntaxKind.EndOfFileToken), typescript_1.default.NodeFlags.None), resultFile);
    return sourceCode;
    function generateRequester({ operationDefinition }) {
        var _a, _b;
        const { verb, path, pathItem, operation, operation: { responses, requestBody }, } = operationDefinition;
        const operationName = getOperationName({ verb, path, operation });
        const returnsJson = apiGen.getResponseType(responses) === 'json';
        let ResponseType = factory_1.factory.createKeywordTypeNode(typescript_1.default.SyntaxKind.UnknownKeyword);
        if (returnsJson) {
            const returnTypes = Object.entries(responses || {})
                .map(([code, response]) => [
                code,
                apiGen.resolve(response),
                apiGen.getTypeFromResponse(response) || factory_1.factory.createKeywordTypeNode(typescript_1.default.SyntaxKind.UndefinedKeyword),
            ])
                .filter(([status, response]) => isDataResponse(status, apiGen.resolve(response), responses || {}))
                .map(([code, response, type]) => typescript_1.default.addSyntheticLeadingComment({ ...type }, typescript_1.default.SyntaxKind.MultiLineCommentTrivia, `* status ${code} ${response.description} `, false))
                .filter((type) => type !== tscodegen_1.keywordType.void);
            if (returnTypes.length > 0) {
                ResponseType = factory_1.factory.createUnionTypeNode(returnTypes);
            }
        }
        const ResponseTypeName = factory_1.factory.createTypeReferenceNode(registerInterface(factory_1.factory.createTypeAliasDeclaration([factory_1.factory.createModifier(typescript_1.default.SyntaxKind.ExportKeyword)], (0, utils_1.capitalize)(operationName + responseSuffix), undefined, ResponseType)).name);
        const parameters = (0, generate_1.supportDeepObjects)([
            ...apiGen.resolveArray(pathItem.parameters),
            ...apiGen.resolveArray(operation.parameters),
        ]);
        const allNames = parameters.map((p) => p.name);
        const queryArg = {};
        for (const param of parameters) {
            const isPureSnakeCase = /^[a-zA-Z][a-zA-Z0-9_]*$/.test(param.name);
            const camelCaseName = (0, lodash_1.camelCase)(param.name);
            let name = isPureSnakeCase && !allNames.includes(camelCaseName) ? camelCaseName : param.name;
            if (name in queryArg) {
                let duplicatedNewName = `${(_a = queryArg[name].param) === null || _a === void 0 ? void 0 : _a.in}_${name}`;
                duplicatedNewName =
                    isPureSnakeCase && !allNames.includes(duplicatedNewName) ? (0, lodash_1.camelCase)(duplicatedNewName) : duplicatedNewName;
                while (duplicatedNewName in queryArg) {
                    duplicatedNewName = '_' + duplicatedNewName;
                }
                queryArg[duplicatedNewName] = queryArg[name];
                queryArg[duplicatedNewName].name = duplicatedNewName;
                delete queryArg[name];
                name = `${param.in}_${name}`;
                name = isPureSnakeCase && !allNames.includes(name) ? (0, lodash_1.camelCase)(name) : name;
            }
            while (name in queryArg) {
                name = '_' + name;
            }
            queryArg[name] = {
                origin: 'param',
                name,
                originalName: param.name,
                type: apiGen.getTypeFromSchema((0, generate_1.isReference)(param) ? param : param.schema),
                required: param.required,
                param,
            };
        }
        if (requestBody) {
            const body = apiGen.resolve(requestBody);
            const schema = apiGen.getSchemaFromContent(body.content);
            const type = apiGen.getTypeFromSchema(schema);
            const schemaName = (0, lodash_1.camelCase)(type.name || (0, generate_1.getReferenceName)(schema) || 'body');
            let name = 'body';
            while (name in queryArg) {
                name = '_' + name;
            }
            for (let media in body.content) {
                body.content[media];
            }
            queryArg[name] = {
                origin: 'body',
                name,
                originalName: schemaName,
                type: getBodyNode(Object.fromEntries(Object.keys(body.content).map((mediaType) => [
                    mediaType,
                    apiGen.getTypeFromSchema(body.content[mediaType].schema),
                ]))),
                required: true,
                body,
            };
        }
        const propertyName = (name) => {
            if (typeof name === 'string') {
                return (0, tscodegen_1.isValidIdentifier)(name) ? factory_1.factory.createIdentifier(name) : factory_1.factory.createStringLiteral(name);
            }
            return name;
        };
        const queryArgValues = Object.values(queryArg);
        const bodyType = (_b = queryArgValues.find((def) => def.origin === 'body')) === null || _b === void 0 ? void 0 : _b.type;
        const getType = () => {
            if (queryArgValues.length === 0) {
                return factory_1.factory.createKeywordTypeNode(typescript_1.default.SyntaxKind.VoidKeyword);
            }
            const queryNode = factory_1.factory.createTypeLiteralNode(queryArgValues
                .filter((def) => def.origin === 'param')
                .map((def) => withQueryComment(factory_1.factory.createPropertySignature(undefined, propertyName(def.name), (0, tscodegen_1.createQuestionToken)(!def.required), def.type), def, true)));
            if (bodyType) {
                return factory_1.factory.createIntersectionTypeNode([queryNode, factory_1.factory.createParenthesizedType(bodyType)]);
            }
            return queryNode;
        };
        const QueryArg = factory_1.factory.createTypeReferenceNode(registerInterface(factory_1.factory.createTypeAliasDeclaration([factory_1.factory.createModifier(typescript_1.default.SyntaxKind.ExportKeyword)], (0, utils_1.capitalize)(operationName + argSuffix), undefined, getType())).name);
        return (0, codegen_1.generateEndpointDefinition)({
            operationName,
            Response: ResponseTypeName,
            QueryArg,
            queryFn: generateQueryFn({ operationDefinition, queryArg }),
        });
    }
    function generateQueryFn({ operationDefinition, queryArg, }) {
        const { path, verb } = operationDefinition;
        const bodyParameter = Object.values(queryArg).find((def) => def.origin === 'body');
        const rootObject = factory_1.factory.createIdentifier('args');
        function pickParams(paramIn) {
            return Object.values(queryArg).filter((def) => def.origin === 'param' && def.param.in === paramIn);
        }
        function createObjectLiteralProperty(parameters, propertyName) {
            return parameters.length === 0
                ? undefined
                : factory_1.factory.createPropertyAssignment(factory_1.factory.createIdentifier(propertyName), factory_1.factory.createObjectLiteralExpression(parameters.map((param) => (0, tscodegen_1.createPropertyAssignment)(param.originalName, accessProperty(rootObject, param.name)), true)));
        }
        return factory_1.factory.createObjectLiteralExpression([
            factory_1.factory.createPropertyAssignment(factory_1.factory.createIdentifier('path'), generatePathExpression(path, pickParams('path'), rootObject, strict)),
            verb.toUpperCase() === 'GET'
                ? undefined
                : factory_1.factory.createPropertyAssignment(factory_1.factory.createIdentifier('method'), factory_1.factory.createStringLiteral(verb.toUpperCase())),
            ...(bodyParameter === undefined
                ? []
                : [
                    factory_1.factory.createPropertyAssignment(factory_1.factory.createIdentifier('body'), factory_1.factory.createPropertyAccessExpression(rootObject, factory_1.factory.createIdentifier(bodyParameter.name))),
                    factory_1.factory.createPropertyAssignment(factory_1.factory.createIdentifier('contentType'), factory_1.factory.createPropertyAccessExpression(rootObject, factory_1.factory.createIdentifier('contentType'))),
                ]),
            createObjectLiteralProperty(pickParams('cookie'), 'cookies'),
            createObjectLiteralProperty(pickParams('header'), 'headers'),
            createObjectLiteralProperty(pickParams('query'), 'params'),
        ].filter(utils_1.removeUndefined), false);
    }
}
exports.generateApi = generateApi;
function accessProperty(rootObject, propertyName) {
    return (0, tscodegen_1.isValidIdentifier)(propertyName)
        ? factory_1.factory.createPropertyAccessExpression(rootObject, factory_1.factory.createIdentifier(propertyName))
        : factory_1.factory.createElementAccessExpression(rootObject, factory_1.factory.createStringLiteral(propertyName));
}
function generatePathExpression(path, pathParameters, rootObject, strict) {
    const expressions = [];
    const head = path.replace(/\{(.*?)\}(.*?)(?=\{|$)/g, (_, expression, literal) => {
        let param = pathParameters.find((p) => p.originalName === expression);
        if (!param) {
            if (strict) {
                throw new Error(`path parameter ${expression} does not seem to be defined in '${path}'!`);
            }
            param = {
                origin: 'param',
                name: expression,
                originalName: expression,
                type: factory_1.factory.createToken(typescript_1.default.SyntaxKind.StringKeyword),
                required: true,
                param: {
                    name: expression,
                    in: 'path',
                    description: 'The name that needs to be deleted',
                    required: true,
                    schema: { type: 'string' },
                },
            };
            pathParameters.push(param);
        }
        expressions.push([param.name, literal]);
        return '';
    });
    return expressions.length
        ? factory_1.factory.createTemplateExpression(factory_1.factory.createTemplateHead(head), expressions.map(([prop, literal], index) => factory_1.factory.createTemplateSpan(accessProperty(rootObject, prop), index === expressions.length - 1
            ? factory_1.factory.createTemplateTail(literal)
            : factory_1.factory.createTemplateMiddle(literal))))
        : factory_1.factory.createNoSubstitutionTemplateLiteral(head);
}
function getBodyNode(bodies) {
    return factory_1.factory.createParenthesizedType(factory_1.factory.createUnionTypeNode(Object.keys(bodies).map((contentType) => factory_1.factory.createTypeLiteralNode([
        factory_1.factory.createPropertySignature(undefined, factory_1.factory.createIdentifier('contentType'), contentType === '*/*' ? factory_1.factory.createToken(typescript_1.default.SyntaxKind.QuestionToken) : undefined, contentType === '*/*'
            ? factory_1.factory.createKeywordTypeNode(typescript_1.default.SyntaxKind.StringKeyword)
            : factory_1.factory.createLiteralTypeNode(factory_1.factory.createStringLiteral(contentType))),
        factory_1.factory.createPropertySignature(undefined, factory_1.factory.createIdentifier('body'), undefined, bodies[contentType]),
    ]))));
}
//# sourceMappingURL=generate.js.map